// Copyright (C) 2023-2026 RabbitMQ Core Team (teamrabbitmq@gmail.com)
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

use crate::test_helpers::*;
use std::error::Error;
#[test]
fn test_list_channels1() -> Result<(), Box<dyn Error>> {
    run_succeeds(["channels", "list"]);

    Ok(())
}

#[test]
fn test_list_channels2() -> Result<(), Box<dyn Error>> {
    run_succeeds(["list", "channels"]);

    Ok(())
}
