// Copyright (C) 2023-2026 RabbitMQ Core Team (teamrabbitmq@gmail.com)
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

mod test_helpers;

mod auth_attempts_tests;
mod bindings_tests;
mod channels_tests;
mod combined_integration_tests;
mod config_file_tests;
mod connections_tests;
mod definitions_export_tests;
mod definitions_import_tests;
mod deprecated_feature_tests;
mod exchange_federation_tests;
mod exchanges_tests;
mod feature_flag_management_tests;
mod feature_flag_tests;
mod federation_link_tests;
mod federation_upstream_uri_modification_tests;
mod health_check_tests;
mod help_tests;
mod memory_breakdown_tests;
mod nodes_tests;
mod operator_policies_tests;
mod pagination_tests;
mod permissions_tests;
mod plugins_tests;
mod policies_conflicting_tests;
mod policies_tests;
mod queue_federation_tests;
mod queues_tests;
mod runtime_parameters_tests;
mod shell_completions_tests;
mod shovel_destination_uri_modification_tests;
mod shovel_source_uri_modification_tests;
mod shovel_tests;
mod streams_tests;
mod test_commands_recommended_against_tests;
mod timeout_tests;
mod tls_tests;
mod user_limits_tests;
mod users_tests;
mod verbose_tests;
mod vhost_limits_tests;
mod vhosts_delete_multiple_tests;
mod vhosts_tests;
