// Copyright (C) 2023-2026 RabbitMQ Core Team (teamrabbitmq@gmail.com)
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#![allow(dead_code)]
#![allow(unused_variables)]

pub(crate) const RABBITMQADMIN_DOC_GUIDE_URL: &str = "https://rabbitmq.com/docs/management-cli";
pub(crate) const RABBITMQ_DOC_GUIDES_URL: &str = "https://rabbitmq.com/docs/";
pub(crate) const GITHUB_DISCUSSIONS_URL: &str =
    "https://github.com/rabbitmq/rabbitmq-server/discussions";
pub(crate) const DISCORD_SERVER_INVITATION_URL: &str = "https://rabbitmq.com/discord/";

pub(crate) const GITHUB_REPOSITORY_URL: &str = "https://github.com/rabbitmq/rabbitmqadmin-ng";

pub(crate) const CONNECTION_GUIDE_URL: &str = "https://rabbitmq.com/docs/connections";
pub(crate) const CHANNEL_GUIDE_URL: &str = "https://rabbitmq.com/docs/channels";
pub(crate) const QUEUE_GUIDE_URL: &str = "https://rabbitmq.com/docs/queues";
pub(crate) const STREAM_GUIDE_URL: &str = "https://rabbitmq.com/docs/streams";
pub(crate) const QUORUM_QUEUE_GUIDE_URL: &str = "https://rabbitmq.com/docs/quorum-queues";
pub(crate) const QUORUM_QUEUE_FAILURE_HANDLING_GUIDE_URL: &str =
    "https://rabbitmq.com/docs/quorum-queues#leader-election";
pub(crate) const TLS_GUIDE_URL: &str = "https://rabbitmq.com/docs/ssl";
pub(crate) const UPGRADE_GUIDE_URL: &str = "https://rabbitmq.com/docs/upgrade";
pub(crate) const BLUE_GREEN_UPGRADE_GUIDE_URL: &str =
    "https://rabbitmq.com/docs/blue-green-upgrade";
pub(crate) const MONITORING_GUIDE_URL: &str = "https://rabbitmq.com/docs/monitoring";
pub(crate) const HEALTH_CHECK_GUIDE_URL: &str =
    "https://rabbitmq.com/docs/monitoring#health-checks";
pub(crate) const FEATURE_FLAG_GUIDE_URL: &str = "https://rabbitmq.com/docs/feature-flags";
pub(crate) const DEPRECATED_FEATURE_GUIDE_URL: &str =
    "https://rabbitmq.com/docs/deprecated-features";
pub(crate) const ACCESS_CONTROL_GUIDE_URL: &str = "https://rabbitmq.com/docs/access-control";
pub(crate) const HTTP_API_ACCESS_PERMISSIONS_GUIDE_URL: &str =
    "https://rabbitmq.com/docs/management#permissions";
pub(crate) const MEMORY_FOOTPRINT_GUIDE_URL: &str = "https://rabbitmq.com/docs/memory-use";
pub(crate) const DEFINITION_GUIDE_URL: &str = "https://rabbitmq.com/docs/definitions";
pub(crate) const CONSUMER_GUIDE_URL: &str = "https://rabbitmq.com/docs/consumers";
pub(crate) const POLLING_CONSUMER_GUIDE_URL: &str = "https://rabbitmq.com/docs/consumers#polling";
pub(crate) const PUBLISHER_GUIDE_URL: &str = "https://rabbitmq.com/docs/publishers";
pub(crate) const CLUSTERING_GUIDE_URL: &str = "https://rabbitmq.com/docs/clustering";
pub(crate) const PEER_DISCOVERY_GUIDE_URL: &str = "https://rabbitmq.com/docs/cluster-formation";
pub(crate) const VIRTUAL_HOST_GUIDE_URL: &str = "https://rabbitmq.com/docs/vhosts";
pub(crate) const VIRTUAL_HOST_LIMIT_GUIDE_URL: &str = "https://rabbitmq.com/docs/vhosts#limits";
pub(crate) const VIRTUAL_HOST_DEFAULT_QUEUE_TYPE_GUIDE_URL: &str =
    "https://rabbitmq.com/docs/vhosts#default-queue-type";
pub(crate) const RUNTIME_PARAMETER_GUIDE_URL: &str = "https://rabbitmq.com/docs/parameters";
pub(crate) const POLICY_GUIDE_URL: &str = "https://rabbitmq.com/docs/policies";
pub(crate) const OPERATOR_POLICY_GUIDE_URL: &str =
    "https://rabbitmq.com/docs/parameters#operator-policies";
pub(crate) const USER_LIMIT_GUIDE_URL: &str = "https://rabbitmq.com/docs/user-limits";
pub(crate) const PASSWORD_GUIDE_URL: &str = "https://rabbitmq.com/docs/passwords";
pub(crate) const PLUGIN_GUIDE_URL: &str = "https://rabbitmq.com/docs/plugins";
pub(crate) const SHOVEL_GUIDE_URL: &str = "https://rabbitmq.com/docs/shovel";
pub(crate) const FEDERATION_GUIDE_URL: &str = "https://rabbitmq.com/docs/federation";
pub(crate) const FEDERATED_QUEUES_GUIDE_URL: &str = "https://rabbitmq.com/docs/federated-queues";
pub(crate) const FEDERATED_EXCHANGES_GUIDE_URL: &str =
    "https://rabbitmq.com/docs/federated-exchanges";
pub(crate) const FEDERATION_REFERENCE_URL: &str = "https://rabbitmq.com/docs/federation-reference";
pub(crate) const COMMERCIAL_OFFERINGS_GUIDE_URL: &str = "https://rabbitmq.com/contact";
